const AWSLambdaCaller = require('./services/AWSLambdaCaller');
const Util = require('./util');
const MediaResolver = require('./util/MediaResolver');
const DialogFlowAccessor = require('./services/DialogFlowAccessor');
const QuoteDataAccessor = require('./services/QuoteDataAccessor');
const KeyMatcherAccessor = require('./services/KeyMatcherAccessor');
const RasaAccessor = require('./services/RasaAccessor');
const RasaLayersAccessor = require('./services/RasaLayersAccessor');
const NLPClassifierAccessor = require('./services/NLPClassifierAccessor');
const ExactMatcherAccessor = require('./services/ExactMatcherAccessor');
const OpenAiAcessor = require('./services/OpenAiAcessor');
const RasaLayersRedisService = require('./services/RasaLayersRedisService');
const defaultFallbackIntent = require('./util/defaultFallbackIntent');
const RequestPreprocessor = require('./util/RequestPreprocessor');
const AsksuiteProcessingNaturalUtils = require('./AsksuiteProcessingNaturalUtils');
const IntEncoder = require('./util/IntEncoder')();
const NodeCache = require('node-cache');
const AsksuiteTranslator = require('./util/Translator');
const StringUtils = require('./util/StringUtils');
const _ = require('lodash');
const PrometheusAcessor = require('./services/PrometheusAccessor');
const GlobalCacheManager = require('./services/GlobalCacheManager');

const DEFAULT_LANGUAGE = 'pt-br';

const GROUP_ID = 'GROUP_ID';
const RASA = 'RASA';
const RASA_LAYERS = 'RASA_LAYERS';
const DIALOG_FLOW = 'DIALOG_FLOW';
const KEY_MATCHER = 'KEY_MATCHER';
const EXACT_MATCHER = 'EXACT_MATCHER';
const OPEN_AI = 'OPEN_AI';
const NLP_CLASSIFIER = 'NLP_CLASSIFIER';
const PROMETHEUS = 'PROMETHEUS';

const intentsNotFoundCache = new NodeCache({ stdTTL: 300, checkperiod: 120 });

const globalCachebleLayers = [{ name: NLP_CLASSIFIER, ttl: 7200 }];

/**
 * @type {GlobalCacheManager}
 */
let globalCacheManager = undefined;

class AsksuiteProcessingNatural {
  static get ORDERED_ACCESSORS() {
    return [
      GROUP_ID,
      RASA,
      RASA_LAYERS,
      EXACT_MATCHER,
      NLP_CLASSIFIER,
      DIALOG_FLOW,
      KEY_MATCHER,
      OPEN_AI,
    ];
  }

  constructor(config) {
    this.config = config;
    this.utils = new AsksuiteProcessingNaturalUtils(this.config);
    this.rasaLayersRedisService = new RasaLayersRedisService({
      config: config.config,
      redis: config.redis,
    });
    this.exactMatcherAccessor = new ExactMatcherAccessor({
      config: config.config,
      redis: config.redisCore || config.redis,
    });

    this.setGlobalCache();
  }

  resolverOrder(config) {
    let resolvers = AsksuiteProcessingNatural.ORDERED_ACCESSORS;

    if (config.USE_PROMETHEUS_LAYER) {
      resolvers = resolvers.map(resolver => (resolver === 'RASA' ? 'PROMETHEUS' : resolver));
    }

    if (!config.RESOLVERS_TO_BE_USED || !config.RESOLVERS_TO_BE_USED.length) {
      return resolvers;
    }

    return resolvers.filter(r => config.RESOLVERS_TO_BE_USED.includes(r));
  }

  findIntent(originalRequest) {
    return new Promise((resolve, reject) => {
      const executor = async originalRequest => {
        let intent = {
          fromCache: false,
        };

        // Formata texto
        const request = await RequestPreprocessor.process(originalRequest);

        const language = this.utils.resolveLanguage(request.language, request.languages);
        const key = `${request.companyId}.${request.userId}.${language}`;

        // Inicializa os conectores
        const awsLambdaCaller = new AWSLambdaCaller(request.config.CORE_LAMBDA_AWS);
        this.dialogFlowAccessor = new DialogFlowAccessor(awsLambdaCaller);
        const keyMatcherAccessor = new KeyMatcherAccessor(
          awsLambdaCaller,
          request.config.OPEN_AI_ENABLED,
        );
        const rasaAccessor = new RasaAccessor(
          request.config.RASA_URL,
          request.config.RASA_CONFIDENCE_THRESHOLD,
        );
        const rasaLayersAccessor = new RasaLayersAccessor(awsLambdaCaller);

        const openAIAccessor = new OpenAiAcessor(
          request.config.OPEN_AI_URL,
          request.config.OPEN_AI_ENABLED,
          originalRequest,
        );

        const nlpClassifierAccessor = new NLPClassifierAccessor(
          request.config.NLP_CLASSIFIER_LAYER_URL,
          request.config.NLP_CLASSIFIER_PARAMETERS,
        );

        const prometheusAcessor = new PrometheusAcessor(
          request.config.PROMETHEUS_PARAMETERS,
          this.config.redis,
        );

        const orderOption = this.resolverOrder(request.config);

        request.findIntent = defaultFallbackIntent(request.companyId);
        request.translated = false;
        request.translation = null;

        const processingSteps = [];

        /**@type {boolean}*/
        const globalCacheEnabled = !!request.config.USE_GLOBAL_CACHE;

        for (const accessorName of orderOption) {
          if (this.shouldUseGlobalCache(globalCacheEnabled, accessorName)) {
            try {
              const cachedIntent = await globalCacheManager.get(accessorName, originalRequest);

              if (cachedIntent) {
                console.log(`OBTAINED ${cachedIntent.intent} BY CACHE FOR LAYER ${accessorName}`);
                intent = cachedIntent;
                break;
              }
            } catch (e) {
              console.log(`ERROR ON GLOBAL CACHE`, e);
            }
          }

          switch (accessorName) {
            case GROUP_ID:
              const extractedValue = IntEncoder.extract(originalRequest.text);
              if (extractedValue) {
                intent = this.utils.findDialogByIntent(request, 'continuar_atendimento');
                const [groupId, attendanceId] = extractedValue.split('-');
                intent.groupId = groupId;
                intent.attendanceId = attendanceId;
              }
              break;
            case DIALOG_FLOW:
              intent = await this.resolveDialogFlowAccessor(request, language);
              break;
            case KEY_MATCHER:
              intent = await keyMatcherAccessor.resolveText(request);
              if (Util.isNaoentendi(intent.intent) && language !== DEFAULT_LANGUAGE) {
                const toSend = await this.utils.copyRequestToDefaultLanguage(request);
                intent = await keyMatcherAccessor.resolveText(toSend);
              }
              intent.keywordTry = { try: true };
              break;
            case RASA:
              intent = await rasaAccessor.resolveText(request, request.needsRasaEntityConfident);
              break;
            case RASA_LAYERS:
              if (request.order && request.order.length) {
                let cached = null;
                try {
                  cached = await this.rasaLayersRedisService.resolveText(
                    request.text,
                    request.language,
                    request.order,
                    request.languages,
                  );
                } catch (e) {
                  console.error(e);
                }
                intent = cached ? cached : await rasaLayersAccessor.resolveText(request);
              }
              break;
            case PROMETHEUS:
              if (request.config.USE_PROMETHEUS_LAYER) {
                const [rasaResult, prometheusResult] = await Promise.all([
                  rasaAccessor.resolveText(request, request.needsRasaEntityConfident),
                  prometheusAcessor.resolveText(
                    request,
                    originalRequest,
                    !!request.config.USE_PROMETHEUS_CACHE,
                  ),
                ]);
                intent = this.getIntentFromPrometheus(rasaResult, prometheusResult);
                break;
              }
              break;
            case EXACT_MATCHER:
              if (request.defaultLanguage) {
                intent = await this.exactMatcherAccessor.resolveText(request);
              }
              break;
            case OPEN_AI:
              intent = await openAIAccessor.resolveText(originalRequest);
              break;
            case NLP_CLASSIFIER:
              if (
                request.config.USE_NLP_CLASSIFIER_LAYER &&
                nlpClassifierAccessor.isSupportedLanguage(language)
              ) {
                const [nlpClassifierResult, dialogFlowResult] = await Promise.all([
                  nlpClassifierAccessor.resolveText(originalRequest),
                  this.resolveDialogFlowAccessor(request, language),
                ]);
                intent = nlpClassifierResult;
                intent.processingStep.dialogFlowResult = dialogFlowResult;
                if (Util.shouldDisambiguateNlpAndDialogFlowResults(intent, dialogFlowResult)) {
                  intent.dialogsToDisambiguate = [
                    { intent: intent.intent, dialog: intent.dialog },
                    { intent: dialogFlowResult.intent, dialog: dialogFlowResult.dialog },
                  ];
                }
              }
              break;
          }
          this.getProcessorStep(intent, accessorName, request.language, processingSteps);
          if (request.findIntent && request.findIntent.groupId) {
            intent.groupId = request.findIntent.groupId;
          }
          request.findIntent = intent;

          if (intent.intent) {
            if (this.shouldUseGlobalCache(globalCacheEnabled, accessorName)) {
              try {
                await globalCacheManager.set(accessorName, originalRequest, intent);
              } catch (err) {
                console.log(`Error setting ${intent} intent into ${accessorName}`, err);
              }
            }
          }

          if (intent.intent && !Util.isNaoentendi(intent.intent)) {
            intent.resolver = accessorName;
            if (intent.resolver === KEY_MATCHER) {
              intent.keywordTry.match = intent.intent;
              intent.keywordTry.intent = intent.intent;
            } else if (intent.resolver === RASA_LAYERS) {
              await this.rasaLayersRedisService
                .saveToCache(
                  intent,
                  request.text,
                  request.language,
                  intent.order,
                  request.languages,
                )
                .catch(console.error);
            }
            //FOR exit condition
            break;
          }
        }

        if (Util.isNaoentendi(intent.intent)) {
          let count =
            originalRequest.subsequentNotUnderstoodCount || intentsNotFoundCache.get(key) || 0;
          intentsNotFoundCache.set(key, ++count);

          const {
            ATTEMPTS_BEFORE_DETECTING_LANGUAGE,
            LANGUAGE_DETECTION_CONFIDENCE,
          } = request.config;

          if (count >= ATTEMPTS_BEFORE_DETECTING_LANGUAGE) {
            const detection = await this.utils.detectLanguage(request);

            if (detection.confidence >= LANGUAGE_DETECTION_CONFIDENCE) {
              const languageCode = detection.isSupported
                ? detection.language.code
                : detection.fallbackLanguage.code;

              if (request.language.substring(0, 2) !== languageCode.substring(0, 2)) {
                intent.languageDetected = detection;
                intentsNotFoundCache.del(key);
              }
            }
          }

          if (count > 1 && !intent.languageDetected && !intent.groupId) {
            const fallbackIntent = defaultFallbackIntent(request.companyId, count);
            intent = this.utils.findDialogByIntent(request, fallbackIntent.intent);
          }
        } else {
          intentsNotFoundCache.del(key);
        }

        intent.translated = request.translated;
        intent.translation = request.translation;
        intent.processingSteps = processingSteps;

        return intent;
      };

      executor(originalRequest)
        .then(resolve)
        .catch(reject);
    });
  }

  getIntentFromPrometheus(rasaResult, prometheusResult) {
    const intent = prometheusResult;
    intent.processingStep.rasaResult = rasaResult;
    return intent;
  }

  async resolveDialogFlowAccessor(request, language) {
    if (!this.dialogFlowResult) {
      request = _.cloneDeep(request);

      this.dialogFlowResult = await this.utils.findIntentDialogFlow(
        this.dialogFlowAccessor,
        request,
        false,
      );

      this.dialogFlowResult.processingStep = this.dialogFlowResult.processingStep || {};

      if (Util.isNaoentendi(this.dialogFlowResult.intent) && language !== DEFAULT_LANGUAGE) {
        this.dialogFlowResult = await this.utils.findIntentDialogFlow(
          this.dialogFlowAccessor,
          request,
          true,
        );
      }

      const processingStep = this.dialogFlowResult.processingStep;
      const isChangeLanguage =
        Util.isChangeLanguage(this.dialogFlowResult.intent) ||
        Util.isChangeLanguage(processingStep.intentFound);

      processingStep.isChangeLanguageRequest = isChangeLanguage;
      if (
        isChangeLanguage &&
        this.dialogFlowResult.parameterFields &&
        this.dialogFlowResult.parameterFields.language
      ) {
        await this.processChangeLanguageRequest(request);
      }
    }
    return this.dialogFlowResult;
  }

  getProcessorStep(intent, resolver, language, processingSteps) {
    if (intent && intent.processingStep) {
      try {
        processingSteps.push({
          resolver,
          language,
          processingStep: JSON.parse(JSON.stringify(intent.processingStep)),
        });

        delete intent.processingStep;
      } catch (e) {
        console.log('[AsksuiteProcessingNatural.getProcessorStep] error', e);
        console.log(
          '[AsksuiteProcessingNatural.getProcessorStep] error intent',
          JSON.stringify(intent),
        );
      }
    }
  }

  resolveMedia(request) {
    return new Promise((resolve, reject) => {
      try {
        const mediaResolver = new MediaResolver(request.companyId);
        mediaResolver
          .resolve(request.media, request.intents)
          .then(resolve)
          .catch(reject);
      } catch (e) {
        reject(e);
      }
    });
  }

  async extractQuoteData(request) {
    request = await RequestPreprocessor.process(request);

    const awsLambdaCaller = new AWSLambdaCaller(request.config.CORE_LAMBDA_AWS);
    const quoteDataAccessor = new QuoteDataAccessor(awsLambdaCaller);

    return await quoteDataAccessor.resolveText(request);
  }

  async extractTextNumbers(request) {
    request = await RequestPreprocessor.process(request);

    const awsLambdaCaller = new AWSLambdaCaller(request.config.CORE_LAMBDA_AWS);
    const quoteDataAccessor = new QuoteDataAccessor(awsLambdaCaller);

    return await quoteDataAccessor.resolveTextNumbers(request);
  }

  async findRasaIntent(request) {
    let intent = {};

    intent = defaultFallbackIntent(request.companyId);
    request.findIntent = intent;

    request = await this.utils.copyRequestToDefaultLanguage(request);
    const rasaAccessor = new RasaAccessor(
      request.config.RASA_URL,
      request.config.RASA_CONFIDENCE_THRESHOLD,
    );

    return await rasaAccessor.resolveText(request, request.needsRasaEntityConfident);
  }

  async processChangeLanguageRequest(request) {
    if (
      !(typeof this.dialogFlowResult.parameterFields === 'object') ||
      !this.dialogFlowResult.parameterFields.language
    ) {
      return;
    }

    this.dialogFlowResult.intent = 'trocar_idioma';

    const languageListValue = this.dialogFlowResult.parameterFields.language.listValue || {
      values: [],
    };

    const googleTranslateKey = request.config.GOOGLE_TRANSLATE_KEY;

    const currentLanguageCode = request.language.substring(0, 2);

    let languageRequested;

    // targetLanguage is used to discover which language the traveller is using
    let targetLanguage = request.translated ? 'pt' : undefined;

    if (!targetLanguage) {
      const detectedLanguage = await AsksuiteTranslator.detectLanguage(
        googleTranslateKey,
        request.text,
      );

      if (detectedLanguage && detectedLanguage.length) {
        targetLanguage = detectedLanguage[0].language;
      } else {
        targetLanguage = currentLanguageCode;
      }
    }

    for (const value of languageListValue.values) {
      const language = StringUtils.unaccent(value.stringValue);

      const valueLanguageInfo = await AsksuiteTranslator.getCodeByLanguage(
        googleTranslateKey,
        language,
        targetLanguage,
      );

      if (valueLanguageInfo && valueLanguageInfo.code != currentLanguageCode) {
        languageRequested = valueLanguageInfo;
        break;
      }
    }

    if (!languageRequested) {
      return;
    }

    const languageInfo = await AsksuiteTranslator.getLanguageByCode(
      googleTranslateKey,
      languageRequested.code,
      languageRequested.code,
    );

    if (!languageInfo) {
      return;
    }

    const enIsSupported = this.utils.isLanguageSupportedByCompany(request, 'en');
    const fallbackLanguageCode = enIsSupported ? 'en' : currentLanguageCode;

    const fallbackLanguage = await AsksuiteTranslator.getLanguageByCode(
      googleTranslateKey,
      fallbackLanguageCode,
      fallbackLanguageCode,
    );
    const isSupported = this.utils.isLanguageSupportedByCompany(request, languageInfo.code);

    if (!isSupported) {
      languageInfo.name = (
        await AsksuiteTranslator.getLanguageByCode(
          googleTranslateKey,
          languageInfo.code,
          fallbackLanguageCode,
        )
      ).name;
    }

    const languageDetected = {
      language: languageInfo,
      isSupported,
      fallbackLanguage,
    };

    this.dialogFlowResult.languageDetected = languageDetected;
    this.dialogFlowResult.processingStep.languageDetected = languageDetected;
  }

  setGlobalCache() {
    if (globalCacheManager) {
      return globalCacheManager;
    }

    if (!this.config.redis) {
      return;
    }

    globalCacheManager = new GlobalCacheManager(this.config.redis);
    globalCacheManager.addLayers(globalCachebleLayers);
  }

  shouldUseGlobalCache(globalCacheEnabled, accessorName) {
    return (
      globalCacheEnabled && globalCacheManager && globalCacheManager.isGlobalLayer(accessorName)
    );
  }
}

module.exports = AsksuiteProcessingNatural;
